@@ -13,10 +13,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import time
-import copy
 import base64
 import hmac
+import time
 import sys
 
 from hashlib import sha1
@@ -36,9 +35,11 @@ from libcloud.utils.xml import fixxpath, findtext
 from libcloud.utils.files import read_in_chunks
 from libcloud.common.types import InvalidCredsError, LibcloudError
 from libcloud.common.base import ConnectionUserAndKey, RawResponse
-from libcloud.common.aws import AWSBaseResponse, AWSDriver, AWSTokenConnection
+from libcloud.common.aws import AWSBaseResponse, AWSDriver, \
+    AWSTokenConnection, SignedAWSConnection
 
 from libcloud.storage.base import Object, Container, StorageDriver
+from libcloud.storage.types import ContainerError
 from libcloud.storage.types import ContainerIsNotEmptyError
 from libcloud.storage.types import InvalidContainerNameError
 from libcloud.storage.types import ContainerDoesNotExistError
@@ -52,9 +53,13 @@ EXPIRATION_SECONDS = 15 * 60
 S3_US_STANDARD_HOST = 's3.amazonaws.com'
 S3_US_WEST_HOST = 's3-us-west-1.amazonaws.com'
 S3_US_WEST_OREGON_HOST = 's3-us-west-2.amazonaws.com'
+S3_CN_NORTH_HOST = 's3.cn-north-1.amazonaws.com.cn'
 S3_EU_WEST_HOST = 's3-eu-west-1.amazonaws.com'
 S3_AP_SOUTHEAST_HOST = 's3-ap-southeast-1.amazonaws.com'
-S3_AP_NORTHEAST_HOST = 's3-ap-northeast-1.amazonaws.com'
+S3_AP_NORTHEAST1_HOST = 's3-ap-northeast-1.amazonaws.com'
+S3_AP_NORTHEAST2_HOST = 's3-ap-northeast-2.amazonaws.com'
+S3_AP_NORTHEAST_HOST = S3_AP_NORTHEAST1_HOST
+S3_SA_EAST_HOST = 's3-sa-east-1.amazonaws.com'
 
 API_VERSION = '2006-03-01'
 NAMESPACE = 'http://s3.amazonaws.com/doc/%s/' % (API_VERSION)
@@ -100,20 +105,9 @@ class BaseS3Connection(ConnectionUserAndKey):
     responseCls = S3Response
     rawResponseCls = S3RawResponse
 
-    def add_default_params(self, params):
-        expires = str(int(time.time()) + EXPIRATION_SECONDS)
-        params['AWSAccessKeyId'] = self.user_id
-        params['Expires'] = expires
-        return params
-
-    def pre_connect_hook(self, params, headers):
-        params['Signature'] = self._get_aws_auth_param(
-            method=self.method, headers=headers, params=params,
-            expires=params['Expires'], secret_key=self.key, path=self.action)
-        return params, headers
-
-    def _get_aws_auth_param(self, method, headers, params, expires,
-                            secret_key, path='/'):
+    @staticmethod
+    def get_auth_signature(method, headers, params, expires, secret_key, path,
+                           vendor_prefix):
         """
         Signature = URL-Encode( Base64( HMAC-SHA1( YourSecretAccessKeyID,
                                     UTF-8-Encoding-Of( StringToSign ) ) ) );
@@ -122,50 +116,34 @@ class BaseS3Connection(ConnectionUserAndKey):
             Content-MD5 + "\n" +
             Content-Type + "\n" +
             Expires + "\n" +
-            CanonicalizedAmzHeaders +
+            CanonicalizedVendorHeaders +
             CanonicalizedResource;
         """
-        special_header_keys = ['content-md5', 'content-type', 'date']
-        special_header_values = {'date': ''}
-        amz_header_values = {}
+        special_headers = {'content-md5': '', 'content-type': '', 'date': ''}
+        vendor_headers = {}
 
-        headers_copy = copy.deepcopy(headers)
-        for key, value in list(headers_copy.items()):
+        for key, value in list(headers.items()):
             key_lower = key.lower()
-            if key_lower in special_header_keys:
-                special_header_values[key_lower] = value.strip()
-            elif key_lower.startswith('x-amz-'):
-                amz_header_values[key.lower()] = value.strip()
-
-        if 'content-md5' not in special_header_values:
-            special_header_values['content-md5'] = ''
-
-        if 'content-type' not in special_header_values:
-            special_header_values['content-type'] = ''
+            if key_lower in special_headers:
+                special_headers[key_lower] = value.strip()
+            elif key_lower.startswith(vendor_prefix):
+                vendor_headers[key_lower] = value.strip()
 
         if expires:
-            special_header_values['date'] = str(expires)
-
-        keys_sorted = list(special_header_values.keys())
-        keys_sorted.sort()
+            special_headers['date'] = str(expires)
 
         buf = [method]
-        for key in keys_sorted:
-            value = special_header_values[key]
+        for _, value in sorted(special_headers.items()):
             buf.append(value)
         string_to_sign = '\n'.join(buf)
 
-        keys_sorted = list(amz_header_values.keys())
-        keys_sorted.sort()
-
-        amz_header_string = []
-        for key in keys_sorted:
-            value = amz_header_values[key]
-            amz_header_string.append('%s:%s' % (key, value))
-        amz_header_string = '\n'.join(amz_header_string)
+        buf = []
+        for key, value in sorted(vendor_headers.items()):
+            buf.append('%s:%s' % (key, value))
+        header_string = '\n'.join(buf)
 
         values_to_sign = []
-        for value in [string_to_sign, amz_header_string, path]:
+        for value in [string_to_sign, header_string, path]:
             if value:
                 values_to_sign.append(value)
 
@@ -175,6 +153,19 @@ class BaseS3Connection(ConnectionUserAndKey):
         )
         return b64_hmac.decode('utf-8')
 
+    def add_default_params(self, params):
+        expires = str(int(time.time()) + EXPIRATION_SECONDS)
+        params['AWSAccessKeyId'] = self.user_id
+        params['Expires'] = expires
+        return params
+
+    def pre_connect_hook(self, params, headers):
+        params['Signature'] = self.get_auth_signature(
+            method=self.method, headers=headers, params=params,
+            expires=params['Expires'], secret_key=self.key, path=self.action,
+            vendor_prefix=self.driver.http_vendor_prefix)
+        return params, headers
+
 
 class S3Connection(AWSTokenConnection, BaseS3Connection):
     """
@@ -227,6 +218,7 @@ class BaseS3StorageDriver(StorageDriver):
     supports_s3_multipart_upload = True
     ex_location_name = ''
     namespace = NAMESPACE
+    http_vendor_prefix = 'x-amz'
 
     def iterate_containers(self):
         response = self.connection.request('/')
@@ -377,10 +369,10 @@ class BaseS3StorageDriver(StorageDriver):
                       'be unique among all the containers in the system',
                 container_name=container_name, driver=self)
         elif response.status == httplib.BAD_REQUEST:
-            raise InvalidContainerNameError(value='Container name contains ' +
-                                            'invalid characters.',
-                                            container_name=container_name,
-                                            driver=self)
+            raise ContainerError(
+                value='Bad request when creating container: %s' %
+                      response.body,
+                container_name=container_name, driver=self)
 
         raise LibcloudError('Unexpected status code: %s' % (response.status),
                             driver=self)
@@ -506,7 +498,7 @@ class BaseS3StorageDriver(StorageDriver):
     def _upload_from_iterator(self, iterator, object_path, upload_id,
                               calculate_hash=True):
         """
-        Uploads data from an interator in fixed sized chunks to S3
+        Uploads data from an iterator in fixed sized chunks to S3
 
         :param iterator: The generator for fetching the upload data
         :type iterator: ``generator``
@@ -602,6 +594,7 @@ class BaseS3StorageDriver(StorageDriver):
 
         if response.status != httplib.OK:
             element = response.object
+            # pylint: disable=maybe-no-member
             code, message = response._parse_error_details(element=element)
             msg = 'Error in multipart commit: %s (%s)' % (message, code)
             raise LibcloudError(msg, driver=self)
@@ -718,18 +711,20 @@ class BaseS3StorageDriver(StorageDriver):
         if delimiter:
             params['delimiter'] = delimiter
 
-        finder = lambda node, text: node.findtext(fixxpath(xpath=text,
-                                                  namespace=self.namespace))
+        def finder(node, text):
+            return node.findtext(fixxpath(xpath=text,
+                                          namespace=self.namespace))
 
         while True:
             response = self.connection.request(request_path, params=params)
 
             if response.status != httplib.OK:
                 raise LibcloudError('Error fetching multipart uploads. '
-                                    'Got code: %s' % (response.status),
+                                    'Got code: %s' % response.status,
                                     driver=self)
 
             body = response.parse_body()
+            # pylint: disable=maybe-no-member
             for node in body.findall(fixxpath(xpath='Upload',
                                               namespace=self.namespace)):
 
@@ -748,6 +743,7 @@ class BaseS3StorageDriver(StorageDriver):
                                         initiator, owner)
 
             # Check if this is the last entry in the listing
+            # pylint: disable=maybe-no-member
             is_truncated = body.findtext(fixxpath(xpath='IsTruncated',
                                                   namespace=self.namespace))
 
@@ -796,7 +792,8 @@ class BaseS3StorageDriver(StorageDriver):
             raise ValueError(
                 'Invalid storage class value: %s' % (storage_class))
 
-        headers['x-amz-storage-class'] = storage_class.upper()
+        key = self.http_vendor_prefix + '-storage-class'
+        headers[key] = storage_class.upper()
 
         content_type = extra.get('content_type', None)
         meta_data = extra.get('meta_data', None)
@@ -804,11 +801,11 @@ class BaseS3StorageDriver(StorageDriver):
 
         if meta_data:
             for key, value in list(meta_data.items()):
-                key = 'x-amz-meta-%s' % (key)
+                key = self.http_vendor_prefix + '-meta-%s' % (key)
                 headers[key] = value
 
         if acl:
-            headers['x-amz-acl'] = acl
+            headers[self.http_vendor_prefix + '-acl'] = acl
 
         request_path = self._get_object_path(container, object_name)
 
@@ -829,7 +826,7 @@ class BaseS3StorageDriver(StorageDriver):
         bytes_transferred = result_dict['bytes_transferred']
         headers = response.headers
         response = response.response
-        server_hash = headers['etag'].replace('"', '')
+        server_hash = headers.get('etag', '').replace('"', '')
 
         if (verify_hash and result_dict['data_hash'] != server_hash):
             raise ObjectHashMismatchError(
@@ -880,10 +877,10 @@ class BaseS3StorageDriver(StorageDriver):
             extra['last_modified'] = headers['last-modified']
 
         for key, value in headers.items():
-            if not key.lower().startswith('x-amz-meta-'):
+            if not key.lower().startswith(self.http_vendor_prefix + '-meta-'):
                 continue
 
-            key = key.replace('x-amz-meta-', '')
+            key = key.replace(self.http_vendor_prefix + '-meta-', '')
             meta_data[key] = value
 
         obj = Object(name=object_name, size=headers['content-length'],
@@ -945,6 +942,16 @@ class S3USWestOregonStorageDriver(S3StorageDriver):
     ex_location_name = 'us-west-2'
 
 
+class S3CNNorthConnection(S3Connection):
+    host = S3_CN_NORTH_HOST
+
+
+class S3CNNorthStorageDriver(S3StorageDriver):
+    name = 'Amazon S3 (cn-north-1)'
+    connectionCls = S3CNNorthConnection
+    ex_location_name = 'CN'
+
+
 class S3EUWestConnection(S3Connection):
     host = S3_EU_WEST_HOST
 
@@ -965,11 +972,46 @@ class S3APSEStorageDriver(S3StorageDriver):
     ex_location_name = 'ap-southeast-1'
 
 
-class S3APNEConnection(S3Connection):
-    host = S3_AP_NORTHEAST_HOST
+class S3APNE1Connection(S3Connection):
+    host = S3_AP_NORTHEAST1_HOST
+
+S3APNEConnection = S3APNE1Connection
 
 
-class S3APNEStorageDriver(S3StorageDriver):
+class S3APNE1StorageDriver(S3StorageDriver):
     name = 'Amazon S3 (ap-northeast-1)'
     connectionCls = S3APNEConnection
     ex_location_name = 'ap-northeast-1'
+
+S3APNEStorageDriver = S3APNE1StorageDriver
+
+
+class S3APNE2Connection(SignedAWSConnection, BaseS3Connection):
+    host = S3_AP_NORTHEAST2_HOST
+    service_name = 's3'
+    version = API_VERSION
+
+    def __init__(self, user_id, key, secure=True, host=None, port=None,
+                 url=None, timeout=None, proxy_url=None, token=None,
+                 retry_delay=None, backoff=None):
+        super(S3APNE2Connection, self).__init__(user_id, key, secure, host,
+                                                port, url, timeout, proxy_url,
+                                                token, retry_delay, backoff,
+                                                4)  # force version 4
+
+
+class S3APNE2StorageDriver(S3StorageDriver):
+    name = 'Amazon S3 (ap-northeast-2)'
+    connectionCls = S3APNE2Connection
+    ex_location_name = 'ap-northeast-2'
+    region_name = 'ap-northeast-2'
+
+
+class S3SAEastConnection(S3Connection):
+    host = S3_SA_EAST_HOST
+
+
+class S3SAEastStorageDriver(S3StorageDriver):
+    name = 'Amazon S3 (sa-east-1)'
+    connectionCls = S3SAEastConnection
+    ex_location_name = 'sa-east-1'
